'''
Copyright (C) 2022, Alejandro Cardenas-Avendano, Alex Lupsasca & Hengrui Zhu
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software Foundation, 
either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. 
If not, see https://www.gnu.org/licenses/.
'''
import numpy as np
import matplotlib.pyplot as plt
import sys
import warnings
import h5py
import os
import imageio.v2 as imageio

#For the analytical calculations
from numpy.lib.scimath import sqrt,  log
from numpy import arccos, arcsin,sin, cos
from scipy.special import ellipk, ellipeinc, ellipe
from scipy.special import ellipkinc as ellipf
from scipy.special import ellipj
from scipy.special import elliprj
from scipy.integrate import cumtrapz,quad

#Required for the wrapper for elliptic integral of the third kind
#import ctypes
#import numpy.ctypeslib as ctl

#For the lensing bands
from scipy.spatial import Delaunay
from matplotlib import path as paths

#Radon transformation
from skimage.transform import radon
from scipy.interpolate import RegularGridInterpolator
from scipy.optimize import curve_fit
from scipy.fft import fft, fftfreq, fftshift
from scipy import interpolate, optimize 

#Plotting 
import matplotlib.ticker as mtick
from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.axes_grid1.inset_locator import zoomed_inset_axes
from mpl_toolkits.axes_grid1.inset_locator import mark_inset
from scipy.spatial import delaunay_plot_2d
import matplotlib
cmap = plt.get_cmap('plasma') # This is the official colors of the BHs!

#Warnings flags
warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
warnings.simplefilter("ignore", UserWarning)
np.seterr(divide='ignore', invalid='ignore')

#Various auxiliary functions
from aart.misc import *

#Elliptic function
from aart.ellipi_f import *

#Lensing Bands functions
import aart.lensingbands as lb

#Raytracing
import aart.raytracing_f as rt

#Illumination Radial Profile
import aart.rprofs_f as ilp

#Observer's Intensity
import aart.intensity_f as obsint

#Visamp
import aart.visamp_f as vamp

#Polarization
from mpmath import polylog as polylog2
import aart.polarization_f as polarizationk


from aart import *
from multiprocessing import get_context
from functools import partial
              
def mp_worker(supergrid0,mask0,sign0, rs0, phi0, t0, supergrid1,mask1,sign1, rs1, phi1, t1, supergrid2,mask2,sign2, rs2, phi2, t2, spin_case, isco, xtend, interpolated3_R, thetao, betaphi, betar, gfactor, sub_kep, N0, N1, N2, tsnap):
    """
    Helper function for implementing 'movie'.
    """
    
    i_bghts0 = obsint.slow_light(supergrid0,mask0,sign0,spin_case,isco,rs0,phi0,np.mod(t0+tsnap,xtend), interpolated3_R,thetao, betaphi, betar, gfactor, sub_kep)
    
    i_bghts1 = obsint.slow_light(supergrid1,mask1,sign1,spin_case,isco,rs1,phi1,np.mod(t1+tsnap,xtend), interpolated3_R,thetao, betaphi, betar, gfactor, sub_kep)
    
    i_bghts2 = obsint.slow_light(supergrid2,mask2,sign2,spin_case,isco,rs2,phi2,np.mod(t2+tsnap,xtend), interpolated3_R,thetao, betaphi, betar, gfactor, sub_kep)
            
    i_I0 = (i_bghts0).reshape(N0,N0).T
    i_I1 = (i_bghts1).reshape(N1,N1).T
    i_I2 = (i_bghts2).reshape(N2,N2).T

    print("Calculating an image at time t=%s (M)"%np.round(tsnap,5))
    return(i_I0,i_I1,i_I2)

def movie(**params): 
    """
    Ray traces a given equatorial source profile, producing several images. Currently, this function is limited to the use of inoisy files (see https://github.com/AFD-Illinois/inoisy). The inoisy file should be saved in the same directory as the file calling this funciton.
    :param spin_case: BH spin
    :param i_case: BH inclination angle, relative to the observer, in degrees
    :param path: path for saving the output. This should be consistent across the usage of all functions in this package
    :param i_frame: inoisy initial time frame for single images
    :param i_fname: sample equatorial profile    
    :param D_obs: observer's distance in units of M  
    :param psi: BH mass-to-distance ratio (default: 1/psi=6.2e9 Kg)
    :param betaphi: angular velocity
    :param betar: radial velocity
    :param gfactor: power of the redshift factor
    :param sub_kep: sub-kepleniarity    
    :param i_tM: initial time in units of M. Makes sense when the time interval is less than the inoisy temporal length
    :param f_tM: final time in units of M. Makes sense when the time interval is less than the inoisy temporal length
    :param snapshots: number of snapshots in range of times [i_tM, f_tM]
    :param nthreads: number of separate flows of execution
    :param limits: limits for the image in units of M. It should coincide with the source profile used for computing and plotting observables 
    """
    
    spins=params.get('spins')
    i_angles=params.get('i_angles')
    path=params.get('path')
    i_frame=params.get('i_frame')
    i_fname=params.get('i_fname')
    D_obs=params.get('D_obs')
    psi=params.get('psi')
    gfactor=params.get('gfactor')
    i_tM=params.get('i_tM')
    f_tM=params.get('f_tM')
    snapshots=params.get('snapshots')
    nthreads=params.get('nthreads')
    limits=params.get('limits')
    bvapp=params.get('bvapp')
    dx0=params.get('dx0')
    dx1=params.get('dx1')
    dx2=params.get('dx2')
    
    for i in range(len(spins)):
        mask=i
        
        spin_case=spins[mask]
        betaphi=params.get('betaphi')[mask]
        betar=params.get('betar')[mask]
        sub_kep=params.get('sub_kep')[mask]
    
        for i_case in i_angles:
  
            isco=rms(spin_case)
            thetao=i_case*np.pi/180

            print("Movies")

            fnbands=path+"LensingBands_a_%s_i_%s_dx0_%s_dx1_%s_dx2_%s.h5"%(spin_case,i_case, dx0, dx1, dx2)

            print("Reading file: ",fnbands)

            h5f = h5py.File(fnbands,'r')

            supergrid0=h5f['grid0'][:]
            mask0=h5f['mask0'][:]
            N0=int(h5f["N0"][0])

            supergrid1=h5f['grid1'][:]
            mask1=h5f['mask1'][:]
            N1=int(h5f["N1"][0])

            supergrid2=h5f['grid2'][:]
            mask2=h5f['mask2'][:]
            N2=int(h5f["N2"][0])

            h5f.close()

            fnrays=path+"Rays_a_%s_i_%s_bv_%s.h5"%(spin_case,i_case, bvapp)

            print("Reading file: ",fnrays)

            h5f = h5py.File(fnrays,'r')

            rs0=h5f['rs0'][:]
            sign0=h5f['sign0'][:]
            t0=h5f['t0'][:]
            phi0=h5f['phi0'][:]

            rs1=h5f['rs1'][:]
            sign1=h5f['sign1'][:]
            t1=h5f['t1'][:]
            phi1=h5f['phi1'][:]

            rs2=h5f['rs2'][:]
            sign2=h5f['sign2'][:]
            t2=h5f['t2'][:]
            phi2=h5f['phi2'][:]

            h5f.close()

            print("Reading inoisy file: ",i_fname)

            hf = h5py.File(i_fname, 'r')

            data = np.array(hf['data/data_env'])
            #inoisy has periodic boudaries, so we need to copy wrap the data with one frame
            data=np.concatenate((data,data[0,:,:][np.newaxis,:,:]),axis=0)

            nt =data.shape[0] #inoisy time resolution
            ni = data.shape[1] #inoisy x resolution
            nj = data.shape[2] #inoisy y resolution

            xtstart = np.array(hf['params/x0start'])[0]
            xtend = np.array(hf['params/x0end'])[0]

            x1start = np.array(hf['params/x1start'])[0]
            x2start = np.array(hf['params/x2start'])[0]
            x1end = np.array(hf['params/x1end'])[0]
            x2end = np.array(hf['params/x2end'])[0]

            x1 = np.linspace(x1start, x1end, ni) 
            x2 = np.linspace(x2start, x2end, nj)

            times = np.linspace(xtstart, xtend, nt)

            h5py.File.close(hf)

            fact=-(D_obs+2*np.log(D_obs))

            t0-=fact
            t1-=fact
            t2-=fact

            fact2=xtend/2-np.nanmax(t0)

            t0+=fact2
            t1+=fact2
            t2+=fact2

            print("AART starts!")

            i_dt = xtend/nt

            Gc=6.67e-11 # G constant [m^3 kg^-1 s^-2]
            cc= 2.99792458e8 # c constant [m/s]
            Msc=1.988435e30 # Solar Mass [Kg]

            MMkg= 6.2e9*psi*Msc # [Kg]
            MM=MMkg *Gc/cc**2 # Mass of the BH in meters, i.e., for M87(psi*6.2*10^9) psi ("Best fit") Solar Masses 

            timeconversion=i_dt*MMkg*Gc/cc**3/(3600*24) # [days]

            interpolated3_R=RegularGridInterpolator((times,x1,x2),  np.transpose(data,(0,2,1)),fill_value=0,bounds_error=False,method='linear')

            I0s = []
            I1s = []
            I2s = []

            func = partial(mp_worker, supergrid0,mask0,sign0, rs0, phi0, t0, supergrid1,mask1,sign1, rs1, phi1, t1, supergrid2,mask2,sign2, rs2, phi2, t2, spin_case, isco, xtend, interpolated3_R, thetao, betaphi, betar, gfactor, sub_kep, N0, N1, N2)

            p = get_context("fork").Pool(nthreads) #using n threads

            if __name__ == 'aart.iMovies':
                I0s,I1s,I2s = zip(*p.map(func, np.linspace(i_tM+i_frame,f_tM,snapshots)))

            p.close()

            filename=path+"Images_a_%s_i_%s_bv_%s_bphi_%s_br_%s_gfact_%s_sk_%s_itM_%s_ftM_%s_shots_%s.h5"%(spin_case,i_case, bvapp, betaphi, betar, gfactor, sub_kep, i_tM, f_tM, snapshots)

            h5f = h5py.File(filename, 'w')
            h5f.create_dataset('bghts0', data=np.array(I0s))
            h5f.create_dataset('bghts1', data=np.array(I1s))
            h5f.create_dataset('bghts2', data=np.array(I2s))
            h5f.create_dataset('tc', data=np.array([timeconversion]))
            h5f.create_dataset('limits', data=np.array([limits]))

            h5f.close()

            print("Images ",filename," created.\n")


from aart import *

def cvisamp(**params):
    """
    Computes the visibility amplitude for specified baesline angles. Requires 'cintensity' and 'clb' from the 'radialintensity' and 'lensingbands' modules respectively to have been called for the corresponding spins and angles.
    :param spins: an iterable object containing BH spins
    :param i_angles: an iterable object containing BH inclination angles, relative to the observer, in degrees
    :param path: path for saving the output. This should be consistent across the usage of all functions in this package
    :param radonangles: iterable object containing the projection angles for the radon transformation
    :param radonfile: takes on values 0 or 1. If equal to 1, the radon cut profiles will be stored
    :param dx0: resolution for the n=0 image in units of M
    :param dx1: resolution for the n=1 image in units of M
    :param dx2: resolution for the n=2 image in units of M
    :param limits: limits for the image in units of M. It should coincide with the source profile used for computing and plotting observables
    :param psi: BH mass-to-distance ratio (default: 1/psi=6.2e9 Kg)
    :param dBH: distance to the BH in meters (default: M87)
    :param fudge: fudge factor (for n>0)
    :param maxbaseline: max baseline in G\lambda
    :param Ncut: Ncut==0 to use the same x-axis baselines for each case
    """
    
    spins=params.get('spins')
    i_angles=params.get('i_angles')
    radonangles=params.get('radonangles')
    radonfile=params.get('radonfile')
    limits=params.get('limits')
    dx0=params.get('dx0')
    dx1=params.get('dx1')
    dx2=params.get('dx2')
    psi=params.get('psi')
    dBH=params.get('dBH')
    maxbaseline=params.get('maxbaseline')
    path=params.get('path')
    Ncut=params.get('Ncut')
    bvapp=params.get('bvapp')
    gfactor=params.get('gfactor')

        
    for i in range(len(spins)):
        mask=i
        
        spin_case=spins[mask]
        betaphi=params.get('betaphi')[mask]
        betar=params.get('betar')[mask]
        gammap=params.get('gammap')[mask]
        sigmap=params.get('sigmap')[mask]
        mup=params.get('mup')[mask]
        sub_kep=params.get('sub_kep')[mask]
        fudge=params.get('fudge')[mask] 
        
        for i_case in i_angles:
            
            isco=rms(spin_case)            
            thetao=i_case*np.pi/180
           
            fnbands=path+"LensingBands_a_%s_i_%s_dx0_%s_dx1_%s_dx2_%s.h5"%(spin_case,i_case, dx0, dx1, dx2)

            print("Reading file: ",fnbands)

            h5f = h5py.File(fnbands,'r')

            supergrid0=h5f['grid0'][:]
            N0=int(h5f["N0"][0])
            supergrid1=h5f['grid1'][:]
            N1=int(h5f["N1"][0])
            supergrid2=h5f['grid2'][:]
            N2=int(h5f["N2"][0])
            h5f.close()

            fint=path+"Intensity_a_%s_i_%s_sk_%s_bphi_%s_br_%s_bvapp_%s_gfact_%s_mup_%s_gp_%s_sp_%s.h5"%(spin_case,i_case, sub_kep, betaphi, betar,  bvapp, gfactor, np.round(mup, 3), gammap, sigmap)

            print("Reading file: ",fint)

            h5f = h5py.File(fint,'r')

            bghts0=h5f['bghts0'][:]
            bghts1=h5f['bghts1'][:]
            bghts2=h5f['bghts2'][:]

            h5f.close()

            vamp.radon_cut(bghts0,bghts1,bghts2,supergrid0,supergrid1,supergrid2, spin_case, i_case, mask, **params)   
           


from aart import *

def raytrace(**params):
    """
    Carries out analytical raytracing and saves the calculations as separate files according to spin-angle combinations.  Requires the 'clb' function in the 'lensingbands' module to have already been called for the correspondsing spins and angles.
    :param spins: an iterable object containing BH spins
    :param i_angles: an iterable object containing BH inclination angles, relative to the observer, in degrees
    :param bvapp: takes on values 0 or 1. If equal to 1, the Beloborodov approximation will also be computed
    :param D_obs: observer's distance in units of M
    :param path: path for saving the output. This should be consistent across the usage of all functions in this package
    """
    
    spins=params.get('spins')
    i_angles=params.get('i_angles')
    bvapp=params.get('bvapp')
    D_obs=params.get('D_obs')
    limits=params.get('limits')
    dx0=params.get('dx0')
    dx1=params.get('dx1')
    dx2=params.get('dx2')
    path=params.get('path')
        
    for a in spins:
        for i in i_angles:
            
            thetao=i*np.pi/180
    
            fnbands=path+"LensingBands_a_%s_i_%s_dx0_%s_dx1_%s_dx2_%s.h5"%(a,i, dx0, dx1, dx2)

            print("Reading file: ",fnbands)

            h5f = h5py.File(fnbands,'r')

            supergrid0=h5f['grid0'][:]
            mask0=h5f['mask0'][:]

            if bvapp!=1:

                supergrid1=h5f['grid1'][:]
                mask1=h5f['mask1'][:]
                supergrid2=h5f['grid2'][:]
                mask2=h5f['mask2'][:]
                h5f.close()

                rt.rt(supergrid0,mask0,supergrid1,mask1,supergrid2,mask2, a, i, **params)
                print("A total of",supergrid0.shape[0]+supergrid1.shape[0]+supergrid2.shape[0],"photons were ray-traced")

            else:

                h5f.close()

                rt.rs_bv(supergrid0,mask0, a, i, **params)
                print("A total of",supergrid0.shape[0],"photons were approximately ray-traced.")
                



from aart import *

# The inverse of the radial potential Eq. 5 P1. 
def OverRint(r,a,lamb,eta):
    return 1/sqrt((r**2 + a**2 - a*lamb)**2 - (r**2 - 2*r + a**2)*(eta + (lamb - a)**2))
    
def ApparentBH(s,a,thetao,alpha,beta,m,sign,distance=10000):

    alpha= s*alpha
    beta= s*beta
    
    # Photon conserved quantities
    # Eqs. (55 P1)
    lamb = -alpha*np.sin(thetao) 
    eta = (alpha**2 - a**2)*np.cos(thetao)**2+beta**2
    
    DeltaTheta = 1/2 *(1 - (eta + lamb**2)/a**2) # Eq. (11 P1)
    
    # Roots of angular potentail
    uP = DeltaTheta + sqrt(DeltaTheta**2 + eta/a**2) # Eq. (11 P1)
    uM = DeltaTheta - sqrt(DeltaTheta**2 + eta/a**2) # Eq. (11 P1)
    
    #Outer horizon of the BH
    rP = 1 + np.sqrt(1 - a**2)

    #We perform the integration Eq. 8a using Eq. 20 P1 in order to get zero. 
    return sqrt(-uM*a**2)*quad(OverRint, rP,distance,args=(a,lamb,eta))[0] +sign*ellipf(np.arcsin(np.cos(thetao)/np.sqrt(uP)), uP/uM) -2*m*ellipk(uP/uM) 

def nlayers(s,a,thetao,thetad,alpha,betai,mbar):
    '''
    Computes the boundary of the nth lensing band
    :param s: Type of ray. s>1 (Rays arrive outside) and s<1 (Rays arrive inside)
              the critical curve. We have to solve for this parameter. 
    :param a: BH's spin (-1<0<1)
    :param thetao: Observers angle [degrees]
    :param thetad: Angle of the disk [degrees]
    :param alpha: Bardeen's coordinate alpha
    :param betai: Bardeen's coordinate beta
    :param mbar: Label of the observed rings 

    :return: 
    '''
    alpha= s*alpha
    beta= s*betai

    #Angular Turning points encountered along the trajectory
    m= mbar + np.heaviside(betai,0) # Eq. (82 P1)

    # Photon conserved quantities
    # Eqs. (55 P1)
    lamb = -alpha*np.sin(thetao) 
    eta = (alpha**2 - a**2)*np.cos(thetao)**2+beta**2
    
    nutheta=np.sign(betai)
    
    # Radial Roots and Integrals 
    AAA = a**2 - eta - lamb**2 # Eq. (79 P3)
    BBB = 2*(eta + (lamb - a)**2) # Eq. (80 P3)
    CCC = -a**2 *eta # Eq. (81 P3)
    
    P = -(AAA**2/12) - CCC # Eq. (85 P3)
    Q = -(AAA/3) *((AAA/6)**2 - CCC) - BBB**2/8  # Eq. (86 P3)

    Delta3 = -4 *P**3 - 27*Q**2 # Eq. (92 P3)
    
    xi0 = np.real(cbrt(-(Q/2) + sqrt(-(Delta3/108))) + cbrt(-(Q/2) - sqrt(-(Delta3/108))) - AAA/3) # Eq. (87 P3)
    z = sqrt(xi0/2) # Eq. (94 P3)
   
    r1 = -z - sqrt(-(AAA/2) - z**2 + BBB/(4*z)) # Eq. (95a P3)
    r2 = -z + sqrt(-(AAA/2) - z**2 + BBB/(4*z)) # Eq. (95b P3)
    r3 = z - sqrt(-(AAA/2) - z**2 - BBB/(4*z))  # Eq. (95c P3)
    r4 = z + sqrt(-(AAA/2) - z**2 - BBB/(4*z))  # Eq. (95d P3)
    
    DeltaTheta = 1/2 *(1 - (eta + lamb**2)/a**2) # Eq. (19 P3)
    
    # Roots of angular potentail
    # Eqs. (B9 P3)
    uP = DeltaTheta + sqrt(DeltaTheta**2 + eta/a**2) # Eq. (19 P3)
    uM = DeltaTheta - sqrt(DeltaTheta**2 + eta/a**2) # Eq. (19 P3)
    
    # Eqs. (B9 P3)
    r21 = r2 - r1 
    r31 = r3 - r1
    r32 = r3 - r2
    r41 = r4 - r1
    r42 = r4 - r2
    r43 = r4 - r3
    
    # Outer and inner horizons
    # Eqs. (2 P3)
    rP = 1 + np.sqrt(1 - a**2)
    rM = 1 - np.sqrt(1 - a**2)
    
    # Eqs. (B10 P3)
    a1=sqrt(-(r43**2/4))
    # Eqs. (B10 P3)
    b1=(r3 + r4)/2
    
    #Elliptic Parameter
    # Eqs. (B13 P3)
    k = (r32*r41)/(r31*r42)

    AA = np.real(sqrt(a1**2 + (b1 - r2)**2)) # Eqs. (B56 P3)
    BB = np.real(sqrt(a1**2 + (b1 - r1)**2)) # Eqs. (B56 P3)

    # This parameter is real and less the unity
    k3 = np.real(((AA + BB)**2 - r21**2)/(4*AA*BB)) # Eqs. (B59 P3)
    
    # Eqs. (20 P2)
    Gtheta = 1/(sqrt(-uM)*a)*(2*m*ellipk(uP/uM) -nutheta*ellipf(np.arcsin(np.cos(thetao)/np.sqrt(uP)), uP/uM) + nutheta*(-1)**m*ellipf(np.arcsin(np.cos(thetad)/np.sqrt(uP)), uP/uM))
    
    if s>1:
        # Eqs.  (A10 P2)
        Q1= 4/sqrt(r31*r42)*ellipf(np.arcsin(sqrt(r31/r41)), k)
        return Q1-Gtheta
    else:
        if k3<1:
            # Eqs. (A11 P2)
            Q2=1/sqrt(AA*BB)*(ellipf(np.arccos((AA - BB)/(AA + BB)), k3) - ellipf(np.arccos((AA *(rP - r1) - BB*(rP - r2))/(AA*(rP - r1) + BB*(rP - r2))), k3))
        else:
            Q2=np.nan
    
        return Q2-Gtheta

def spacedmarks(x, y, Nmarks):

    """
    Computes the arch-length
    :param x: x point
    :param y: y point
    :param Nmarks: Number of marks 

    :returns: position of the x and y markers
    """
    dydx = np.gradient(y, x[0],edge_order=2)
    dxdx = np.gradient(x, x[0],edge_order=2)
    arclength = cumtrapz(sqrt(dydx**2 + dxdx**2), initial=0)
    marks = np.linspace(0, max(arclength), Nmarks)
    markx = np.interp(marks, arclength, x)
    marky = np.interp(markx, x, y)
    return markx, marky
    

def CritCurve(a,angle):
    """
    Computes the critical curve (aka the black hole shadow)
    :param a: spin of black hole
    :param angle: angle of the observer in degrees

    :returns: contour of the critical curve on the observer plane
    """
    thetao = angle * np.pi/180
    
    #Critical roots of the radial potential (Eq. 5 P2) given by (Eq. 40 P2)
    rM = 2*(1 + np.cos(2/3 *np.arccos(-(a))))
    rP = 2*(1 + np.cos(2/3 *np.arccos(a)))
    
    #Define the region where \tilde{r} lies
    r=np.linspace(rM,rP,int(1e7))
    
    #Eqs. 38 & 39 P2
    lam = a + r/a *(r - (2 *(r**2 - 2*r + a**2))/(r - 1))
    eta = r**3/a**2 *((4*(r**2 - 2*r + a**2))/(r - 1)**2 - r)

    #Eqs. 55 P2
    alpha=-lam/np.sin(thetao)
    # This is beta squared actually, but we need to check where it is positive
    beta=eta + a**2 *np.cos(thetao)**2 - lam**2*np.tan(thetao)**(-2)
    
    mask=np.where(beta>0)
    r=r[mask]
    
    rmin=min(r)+1e-12
    rmax=max(r)-1e-12
    
    r=np.linspace(rmin,rmax,int(1e6))
        
    #Eqs. 38 & 39 P2
    lam = a + r/a *(r - (2 *(r**2 - 2*r + a**2))/(r - 1))
    eta = r**3/a**2 *((4*(r**2 - 2*r + a**2))/(r - 1)**2 - r)

    #Eqs. 55 P2
    alpha=-lam/np.sin(thetao)
    beta=eta + a**2 *np.cos(thetao)**2 - lam**2*np.tan(thetao)**(-2)

    return alpha, sqrt(beta)

def round_up_to_even(f):
    return int(np.ceil(f / 2.) * 2)

def in_hull(p, hull):
    """
    Check if points in p are inside the concave hull
    """
    concave=paths.Path(hull)
    return concave.contains_points(p)

def grid_mask(hull,hull2,dx,limits,force_lims = False):
    """
    create cartesian grid on the observer plane
    :param hull: marks outer edge of the lensing band
    :param hull2: marks inner edge of the lensing band
    :param dx: grid resolution
    :param limits: specify limits for the both axis on the observer plane, only works if force_lims = True
    :param force_lims: specify a limit or find the optimal limit for the grid

    :returns: a cartesian grid and a mask indicating lensing band, along with other parameters
    """
    
    if force_lims == False:
        xlim_min = hull2[:,0].min()-5*dx
        xlim_max = hull2[:,0].max()+5*dx
        ylim_min = hull2[:,1].min()-5*dx
        ylim_max = hull2[:,1].max()+5*dx
        lims = np.ceil(max(np.abs(xlim_min),np.abs(xlim_max),np.abs(ylim_min),np.abs(ylim_max)))
        if lims>=limits:
            lims = limits
        x = np.linspace(-lims, lims, round_up_to_even(2*lims/dx))
        y = np.linspace(-lims, lims, round_up_to_even(2*lims/dx))
        N = int(round_up_to_even(2*lims/dx))
    else:
        lims = limits
        x = np.linspace(-limits, limits, round_up_to_even(2*limits/dx))
        y = np.linspace(-limits, limits, round_up_to_even(2*limits/dx))
        N = int(round_up_to_even(2*limits/dx))
    mesh = np.array(np.meshgrid(x , y))
    grid=mesh.T.reshape(-1, 2)
    
    # We check the points belong to the lensing bands
    mask1=np.invert(in_hull(grid,hull))
    mask2=in_hull(grid,hull2)

    indexes=mask2*mask1

    return grid, N , indexes, lims

def hulls(alpha, beta,limits, spin_case, thetao, thetad, D_obs, smin=0.5, smax=100,limi0=0.99,lime0=1.01,limi1=0.999,lime1=1.001,limi2=0.9999,lime2=1.001):

    data=(np.append(alpha,alpha[::-1]), np.append(beta,-beta[::-1]))

    points_0i=np.zeros([data[0].size,2])
    points_0e=np.array([[-limits,-limits],[limits,-limits],[limits,limits],[-limits, limits]])

    points_1i=np.zeros([data[0].size,2])
    points_1e=np.zeros([data[0].size,2])

    points_2i=np.zeros([data[0].size,2])
    points_2e=np.zeros([data[0].size,2])

    for i in range(data[0].size):
        if data[1][i]>=0:
            m1=optimize.root(ApparentBH, limi0, args=(spin_case,thetao,data[0][i],data[1][i],1,1,D_obs))
        else:
            m1=optimize.root(ApparentBH, limi0, args=(spin_case,thetao,data[0][i],data[1][i],0,-1,D_obs))

        points_0i[i]=m1.x[0]*np.array([data[0][i],data[1][i]])
        
        m1=optimize.root(nlayers, limi1, args=(spin_case,thetao,thetad,data[0][i],data[1][i],1))
        m2=optimize.root(nlayers, lime1, args=(spin_case,thetao,thetad,data[0][i],data[1][i],1))

        points_1i[i]=limi1*m1.x[0]*np.array([data[0][i],data[1][i]])
        points_1e[i]=lime1*m2.x[0]*np.array([data[0][i],data[1][i]])

        if m1.x[0]<smin:
            points_1i[i]=limi1*smin*np.array([data[0][i],data[1][i]])
        else:
            points_1i[i]=limi1*m1.x[0]*np.array([data[0][i],data[1][i]]) 

        if m2.x[0]>smax:
            points_1e[i]=lime1*smax*np.array([data[0][i],data[1][i]])
        else:
            points_1e[i]=lime1*m2.x[0]*np.array([data[0][i],data[1][i]])

        m1=optimize.root(nlayers, limi2, args=(spin_case,thetao,thetad,data[0][i],data[1][i],2))
        m2=optimize.root(nlayers, lime2, args=(spin_case,thetao,thetad,data[0][i],data[1][i],2))

        points_2i[i]=limi2*m1.x[0]*np.array([data[0][i],data[1][i]])
        points_2e[i]=lime2*m2.x[0]*np.array([data[0][i],data[1][i]])

    return points_0i, points_0e, points_1i, points_1e, points_2i,  points_2e


def lb(spin_case, i_case, **params):
    
    npointsS=params.get('npointsS')
    limits=params.get('limits')
    thetad=params.get('thetad')
    D_obs=params.get('D_obs')
    p_image=params.get('p_image')
    dx0=params.get('dx0')
    dx1=params.get('dx1')
    dx2=params.get('dx2')
    path=params.get('path')
    
    #Observer's inclination in radians
    thetao=i_case*np.pi/180

    #"Computing the grids for each lensing bands"
    critc=CritCurve(spin_case,i_case)

    alpha_critc, beta_critc = spacedmarks(critc[0], critc[1], npointsS)

    hull_0i, hull_0e, hull_1i, hull_1e, hull_2i, hull_2e = hulls(alpha_critc, beta_critc, limits, spin_case, thetao, thetad, D_obs)

    if p_image==1:
        supergrid0, N0, mask0, lim0 =grid_mask(hull_0i,hull_0e,dx0,limits,True)
        print("Number of points in the n=0 grid ", supergrid0.shape[0])

        supergrid1, N1, mask1, lim1 =grid_mask(hull_1i,hull_1e,dx1,limits,True)
        print("Number of points in the n=1 grid ", supergrid1.shape[0])

        supergrid2, N2, mask2, lim2 =grid_mask(hull_2i,hull_2e,dx2,limits,True)
        print("Number of points in the n=2 grid ", supergrid2.shape[0])
        
    else:
        supergrid0, N0, mask0, lim0 =grid_mask(hull_0i,hull_0e,dx0,limits)
        print("Number of points in the n=0 grid ", supergrid0.shape[0])

        supergrid1, N1, mask1, lim1 =grid_mask(hull_1i,hull_1e,dx1,limits)
        print("Number of points in the n=1 grid ", supergrid1.shape[0])

        supergrid2, N2, mask2, lim2 =grid_mask(hull_2i,hull_2e,dx2,limits)
        print("Number of points in the n=2 grid ", supergrid2.shape[0])
        
    # Create a directory for the results
    isExist = os.path.exists(path)
    if not isExist:
        os.makedirs(path)
        print("A directory (Results) was created to store the results")

    filename=path+"LensingBands_a_%s_i_%s_dx0_%s_dx1_%s_dx2_%s.h5"%(spin_case,i_case, dx0, dx1, dx2)
    h5f = h5py.File(filename, 'w')

    h5f.create_dataset('alpha', data=alpha_critc)
    h5f.create_dataset('beta', data=beta_critc)

    h5f.create_dataset('hull_0i', data=hull_0i)
    h5f.create_dataset('hull_0e', data=hull_0e)
    h5f.create_dataset('grid0', data=supergrid0)
    h5f.create_dataset('N0', data=np.array([N0]))
    h5f.create_dataset('mask0', data=mask0)
    h5f.create_dataset('lim0', data=np.array([lim0]))

    h5f.create_dataset('hull_1i', data=hull_1i)
    h5f.create_dataset('hull_1e', data=hull_1e)
    h5f.create_dataset('grid1', data=supergrid1)
    h5f.create_dataset('N1', data=np.array([N1]))
    h5f.create_dataset('mask1', data=mask1)
    h5f.create_dataset('lim1', data=np.array([lim1]))

    h5f.create_dataset('hull_2i', data=hull_2i)
    h5f.create_dataset('hull_2e', data=hull_2e)
    h5f.create_dataset('grid2', data=supergrid2)
    h5f.create_dataset('N2', data=np.array([N2]))
    h5f.create_dataset('mask2', data=mask2)
    h5f.create_dataset('lim2', data=np.array([lim2]))

    h5f.close()

    print("File ",filename," created.")
    
def clb(**params):
    
    """
    Computes the grids for each of the lensing bands and saves them as separate files according to spin-angle combinations.
    :param spins: an iterable object containing BH spins
    :param i_angles: an iterable object containing BH inclination angles, relative to the observer, in degrees
    :param npointsS: number of points in the critical curve
    :param limits: limits for the image in units of M. It should coincide with the source profile used for computing and plotting observables
    :param D_obs: observer's distance in units of M
    :param p_image: takes on values 0 or 1. If equal to 1, the sizes of the grids will be equal and an image can be computed
    :param dx0: resolution for the n=0 image in units of M
    :param dx1: resolution for the n=1 image in units of M
    :param dx2: resolution for the n=2 image in units of M
    :param path: path for saving the output. This should be consistent across the usage of all functions in this package
    """
    spins=params.get('spins')
    i_angles=params.get('i_angles')
    
    for a in spins:
        for i in i_angles:
            lb(a, i, **params)
            
    


from aart import *

# default form of Johnson's SU distribution parameter mu_p
def mu_p(spin_case):
    return 1-np.sqrt(1-spin_case**2)

def generate_dict(spins, i_angles, fudge, dBH=5.214795112e23, psi=1.07473555940836, D_obs=10000, sub_kep=[1.0], betar=[1.0], betaphi=[1.0], radonfile=0, bvapp=0, p_image=1, limits=25, dx0=0.02, dx1=0.02, dx2=0.02, radonangles=[0,90], i_fname="inoisy.h5", disk="dynamical", i_frame=0, i_tM=0, f_tM=12, snapshots=12, mup=[1.0], gammap=[-3/2], sigmap=[1/2], gfactor=3, maxbaseline=500, npointsS=100, nthreads=4, Ncut=0, path = './Results/'):
     
    """
    Produces a dictionary containing all the parameters for the simulations.
    :param spins: an iterable object containing BH spins
    :param i_angles: an iterable object containing BH inclination angles, relative to the observer, in degrees
    :param fudge: array of fudge factors (for n>0), indexed according to spins array
    :param dBH: distance to the BH in meters (default: M87)
    :param psi: BH mass-to-distance ratio (default: 1/psi=6.2e9 Kg)
    :param D_obs: observer's distance in units of M
    :param sub_kep: array of sub-kepleniarity values, indexed according to spins array
    :param betar: array of radial velocity values, indexed according to spins array
    :param betaphi: array of angular velocity values, indexed according to spins array
    :param radonfile: takes on values 0 or 1. If equal to 1, the radon cut profiles will be stored
    :param bvapp: takes on values 0 or 1. If equal to 1, the Beloborodov approximation will be computed
    :param p_image: takes on values 0 or 1. If equal to 1, the sizes of the grids will be equal and an image can be computed
    :param limits: limits for the image in units of M. It should coincide with the source profile used for computing and plotting observables
    :param dx0: resolution for the n=0 image in units of M
    :param dx1: resolution for the n=1 image in units of M
    :param dx2: resolution for the n=2 image in units of M   
    :param radonangles: iterable object containing the projection angles for the radon transformation
    :param i_fname: sample equatorial profile    
    :param disk: takes on values "stationary", which assumes a single inoisy frame, or "dyamical"
    :param i_frame: inoisy initial time frame for single images
    :param i_tM: initial time in units of M. Makes sense when the time interval is less than the inoisy temporal length
    :param f_tM: final time in units of M. Makes sense when the time interval is less than the inoisy temporal length
    :param snapshots: number of snapshots in range of times [i_tM, f_tM]
    :param gammap: array of values for Johnson's SU distribution parameter, indexed according to spins array
    :param mup: array of values for Johnson's SU distribution parameter, indexed according to spins array
    :param sigmap: array of values for Johnson's SU distribution parameter, indexed according to spins array
    :param gfactor: power of the redshift factor
    :param maxbaseline: max baseline in G\lambda
    :param npointsS: number of points in the critical curve
    :param nthreads: number of separate flows of execution  
    :param Ncut: Ncut==0 to use the same x-axis baselines for each case
    :param path: path for saving the output. This should be consistent across the usage of all functions in this package
    """
    

    
    #Disk's inclination  
    #Current version is just implemented for equatorial models   
    i_disk=90    
    thetad=i_disk*np.pi/180
    
#     # constants
#     Gc=6.67e-11 # G constant [m^3 kg^-1 s^-2]
#     cc= 2.99792458e8 # c constant [m/s]
#     Msc=1.988435e30 # Solar Mass [Kg]

#     MMkg= 6.2e9*psi*Msc # [Kg]
#     MM=MMkg *Gc/cc**2 # Mass of the BH in meters, i.e., for M87(psi*6.2*10^9) psi ("Best fit") Solar Masses 

#     # Size of the real image in meters
#     sizeim_Real=(limits)*MM 
#     #1 microarcsec in radians
#     muas_to_rad = np.pi/648000 *1e-6 
#     fov_Real=np.arctan(sizeim_Real/(dBH))/muas_to_rad #muas
    
    params = {
        "spins": spins,
        "i_angles": i_angles,
        "fudge": fudge,
        "dBH": dBH,
        "psi":psi,
        "D_obs": D_obs,
        "sub_kep": sub_kep,
        "betar": betar,
        "betaphi": betaphi,
        "radonfile": radonfile,
        "bvapp": bvapp,
        "p_image": p_image,
        "limits": limits,
        "dx0": dx0,
        "dx1": dx1,
        "dx2": dx2, 
        "radonangles": radonangles,
        "i_fname": i_fname,
        "disk": disk,
        "i_frame": i_frame,
        "i_tM": i_tM,
        "f_tM": f_tM,
        "snapshots": snapshots,
        "gammap": gammap,
        "mup": mup,
        "sigmap": sigmap,
        "gfactor": gfactor,
        "maxbaseline": maxbaseline,
        "npointsS": npointsS,
        "nthreads": nthreads, 
        "path": path,
        "thetad": thetad,
        "Ncut":Ncut
    }
    return params



import numpy as np
import h5py
from scipy.signal import find_peaks
from scipy.interpolate import CubicSpline
from scipy.optimize import curve_fit
from scipy.signal import savgol_filter
from aart import *
import linecache
import sys

## defining constants for unit conversions (makes sin functinon dimensionless when fitting to V_fit)

dM=5.214795112e23  # Distance to M87 in meters
psi=1.07473555940836 # Mass of M87 1 /psi= 6.2e9 Kg
Gc=6.67e-11 # G constant [m^3 kg^-1 s^-2]
cc= 2.99792458e8 # c constant [m/s]
Msc=1.988435e30 # Solar Mass [Kg]
MMkg= 6.2e9*psi*Msc # [Kg]
MM87=MMkg *Gc/cc**2 # Mass of M87 in meters, i.e., (psi*6.2*10^9) psi ("Best fit") Solar Masses 
sizeim=MM87 # Size of M in meters
muas_to_rad = np.pi/648000 *1e-6 #1 microarcsec in radians
unitfact=1/(muas_to_rad*1e9)
M_to_muas=np.arctan(sizeim/(dM))/muas_to_rad

def PrintException():
    exc_type, exc_obj, tb = sys.exc_info()
    f = tb.tb_frame
    lineno = tb.tb_lineno
    filename = f.f_code.co_filename
    linecache.checkcache(filename)
    line = linecache.getline(filename, lineno, f.f_globals)
    print('EXCEPTION IN ({}, LINE {} "{}"): {}'.format(filename, lineno, line.strip(), exc_obj))
    

## gets calculated visibility amplitudes stored in directory 'path'
def get_visamps(mask1, mask2, **params):
    spin_case=params.get('spins')[mask1]
    i_case=params.get('i_angles')[mask2]
    
    radonangles=params.get('radonangles')
    path=params.get('path')
    radonfile=params.get('radonfile')
    dx0=params.get('dx0')
    dx1=params.get('dx1')
    dx2=params.get('dx2')
    Ncut=params.get('Ncut')    
    gfactor=params.get('gfactor')
    fudge=params.get('fudge')[mask1]
    fudge=params.get('fudge')[mask1]
    bvapp=params.get('bvapp')
    betaphi=params.get('betaphi')[mask1]
    betar=params.get('betar')[mask1]
    gfactor=params.get('gfactor')
    gammap=params.get('gammap')[mask1]
    sigmap=params.get('sigmap')[mask1]
    path=params.get('path')
    mup=params.get('mup')[mask1]
    sub_kep=params.get('sub_kep')[mask1]
    fudge=params.get('fudge')[mask1]
    
    visamps = []

    for i in range(len(radonangles)):
        radonangle=radonangles[i]
        fnvisamps=path+"Visamp_%s_a_%s_i_%s_dx0_%s_dx1_%s_dx2_%s_sk_%s_bphi_%s_br_%s_bvapp_%s_gfact_%s_gp_%s_sp_%s_fudge_%s_Ncut_%s.h5"%(radonangle,spin_case,i_case, dx0, dx1, dx2, sub_kep, betaphi, betar,  bvapp, gfactor, gammap, sigmap, fudge, int(Ncut))

        h5f = h5py.File(fnvisamps,'r')
        
        '''
        if Ncut==0:
            freqs=h5f['freqs'][:]
        '''    
        visamps.append(h5f['visamp'][:])
        
        h5f.close()
        
#     return np.array(visamps), np.array(norms)  
    return np.array(visamps)

## gets frequencies used in visbility amplitude file stored in directory 'path'
def get_freqs(mask1, mask2, **params):
    
    spin_case=params.get('spins')[mask1]
    i_case=params.get('i_angles')[mask2]
    radonangles=params.get('radonangles')
    path=params.get('path')
    radonfile=params.get('radonfile')
    dx0=params.get('dx0')
    dx1=params.get('dx1')
    dx2=params.get('dx2')
    Ncut=params.get('Ncut')    
    gfactor=params.get('gfactor')
    fudge=params.get('fudge')[mask1]
    bvapp=params.get('bvapp')
    betaphi=params.get('betaphi')[mask1]
    betar=params.get('betar')[mask1]
    gfactor=params.get('gfactor')
    gammap=params.get('gammap')[mask1]
    sigmap=params.get('sigmap')[mask1]
    path=params.get('path')
    mup=params.get('mup')[mask1]
    sub_kep=params.get('sub_kep')[mask1]
    fudge=params.get('fudge')[mask1]
    fnvisamps=path+"Visamp_%s_a_%s_i_%s_dx0_%s_dx1_%s_dx2_%s_sk_%s_bphi_%s_br_%s_bvapp_%s_gfact_%s_gp_%s_sp_%s_fudge_%s_Ncut_%s.h5"%(radonangles[0],spin_case,i_case, dx0, dx1, dx2, sub_kep, betaphi, betar,  bvapp, gfactor, gammap, sigmap, fudge, int(Ncut))

    h5f = h5py.File(fnvisamps,'r')
    
    freqs=h5f['freqs'][:]

    h5f.close()
        
    return freqs

## returns points (x, y) corresponding to maxima in y, with a y value greater than minval
def filter_maxima(x, y, minval):
    x_new = []
    y_new = []
    for k in range(1, len(x)-1):
            if (y[k+1]-y[k]<0) and (y[k]-y[k-1]>0):
                if y[k]>minval:
                    x_new.append(x[k])
                    y_new.append(y[k])
    return x_new, y_new

## finds argument of element in 'array' with value closest to 'value'
def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx

## given arrays of arrays x and y, returns a set of curves beginning at each element in x[0], where the next point in the ith curve is that in x[1] closest to x[0][i], and so on
def generate_curves(x, y):
    x_curves = []   ## x-values for the curves
    y_curves = []   ## y-values for the curves
        
    Lx = len(x)
    
    for j in range(0,len(x[0])):
        x_curve=[x[0][j]]
        y_curve=[y[0][j]]
        
        x_new=x_curve
        for i in range(1,Lx):
            closest=find_nearest(x[i], x_new)
            x_curve.append(x[i][closest])
            y_curve.append(y[i][closest])
            x_new=x[i][closest]
            
        x_curves.append(x_curve)
        y_curves.append(y_curve)
    return x_curves, y_curves

def circlipse(x, r0, r1, r2, phi0):
        return 2*r0 + 2*np.sqrt((r1**2)*(np.sin(x-phi0)**2)+(r2**2)*(np.cos(x-phi0)**2))
    
## takes in a clean 1d signal x, and outputs a corresponding signal with gaussian noise profile (mean=mu, std=s)
def gaussian_noise(x, mu, s):
    noise = np.random.normal(mu,s,len(x))
    #return x + noise
    return np.abs(x + noise)

## define a function to add noise to each snapshot of the visibility spectrum and average into one profile
def average_snapshots(visamps, mu_n, s_n, norms):
    """
    :param norms: nested array containing 36 norms for each snapshot in the ns=0.0 visamp file corresponding to "visamps"
    """
    noisy_visamps=[]
    normmax=np.max([np.max(norms[i]) for i in range(len(norms))])
    for i in range(len(visamps)):
        ## normalize so that the zero astro noise profile at snaphshot i starts at 1 for all baseline angles
        visampss=visamps[i]/normmax

        ## add noise to each visibility profile across the 36 baseline angles
        visampss=np.array([rf.gaussian_noise(vis, mu_n, s_n) for vis in visampss])
        noisy_visamps.append(visampss)

    ## now average the snapshots that now have both instrumental and astrophysical noise
    visamps_n=np.zeros(noisy_visamps[0].shape) 
    for i in range(len(noisy_visamps)):
        visamps_n+=noisy_visamps[i]/len(noisy_visamps)
    return visamps_n, normmax

## given a visibility profile across specified baseline angles (=radonangles), stores its radial profile in directory 'path'
def radialfit(visamps, freq, radonangles, mu_n, s_n, ulim1_fit, ulim2_fit, gfit_min, path, p0_ratio, norms=[]):
    """
    :param visamps: 2d array of visibility amplitdues values for different radonangles
    :param freq: array of baseline frequencies over which each visibility amplitude is recorded. 
    :param radonangles: iterable object containing the projection angles for the radon transformation
    :param mu_n: mean of the gaussian noise added to visamps
    :param s_n: standard deviation of the gaussian noise added to visamps
    :param ulim1_fit: beginning of the baseline window for fits
    :param ulim2_fit: end of the baseline window for fits
    :param gfit_min: minimum goodness of fit for projected diameter in fits
    :param path: path for saving the output
    :param p0_ratio: parameter in (0,1) used in intial guess p0=[aL_max*p0_ratio, aR_max*p0_ratio, 40] for curve_fit 
    """
    
    if len(norms)!=0:
        normax=np.max(norms)
        visamps=visamps/normax

    mask=np.where((freq>=ulim1_fit) & (freq<=ulim2_fit))[0]
    freqfit=freq[mask]
    datas=[]
    power=np.array([])
    upower=(ulim2_fit+ulim1_fit)/2
    datas=np.array([gaussian_noise(visamps[i], mu_n, s_n) for i in range (len(radonangles))])
    d_vals=[]
    g_vals=[]
    
    try:
        for i in range(len(radonangles)):
            data=datas[i][mask]

            ## in presence of noise, aL and aR are fixed parameters
            def visamp_func(u, aL, aR, dphi):
                return np.sqrt(aL**2+aR**2+2*aL*aR*np.sin(2*np.pi*dphi*u/unitfact))/np.sqrt(u*1e9)

            ## goodness of fit
            def g_fit(aL, aR, d, x, y):
                deviation= np.sum((visamp_func(x, aL, aR, d)-np.abs(y))**2)
                avg_v_fit = np.sum(visamp_func(x, aL, aR, d))
                return np.exp(-np.sqrt(deviation/len(x))/(avg_v_fit/len(x)))

            ## approximate bounds for aL and aR
            visamp=data
            vmax=np.max(visamp)
            umax=freqfit[np.argmax(visamp)]
            vmin=np.min(visamp)
            umin=freqfit[np.argmin(visamp)]

            aL_max=(1/2)*(vmax*np.sqrt(1e9*umax)+vmin*np.sqrt(1e9*umin))
            aR_max=(1/2)*(vmax*np.sqrt(1e9*umax)-vmin*np.sqrt(1e9*umin))


            ## find best fit params
            bounds = ([0,0,35], [aL_max, aL_max, 45]) 
            p0 = [aL_max*p0_ratio, aR_max*p0_ratio, 40]
            params, pcov_v = curve_fit(visamp_func, freqfit, data,p0=p0, bounds=bounds, maxfev=2000)  

            aL0 = params[0]
            aR0 = params[1]
            dphi0 = params[2]

            drange = np.arange(8*M_to_muas, 12*M_to_muas, 1/100)

            ## corresponding goodness of fit for each d \in drange
            gvals = [g_fit(aL0, aR0, d, freqfit, data) for d in drange]

            ## keep values of d with g(d) above gfit_min
            d_maxima, g_maxima = filter_maxima(drange, gvals, gfit_min)
            
            d_vals.append(np.array(d_maxima))
            g_vals.append(np.array(g_maxima))

        d_vals=np.array(d_vals)
        g_vals=np.array(g_vals)
        
        # construct circlipses  
        circlipses, g_circlipses = generate_curves(d_vals, g_vals) ## generate various circlipses
        g_circlipses = [np.prod(g) for g in g_circlipses]  ## joint goodness of fit for each circlipse

        best_circlipse = circlipses[np.array(g_circlipses).argmax()]


        params, pcov_v = curve_fit(circlipse, np.array(radonangles)*np.pi/180, best_circlipse,bounds=([0,0,0,-np.pi/2],[50,50,50,np.pi/2]),maxfev=2000)
        

        RMSD_a=np.sqrt(np.average((best_circlipse-circlipse(np.array(radonangles)*np.pi/180,*params))**2))/np.average(circlipse(np.array(radonangles)*np.pi/180,*params))

        RMSD_b=np.sqrt(np.average((best_circlipse-circlipse(np.array(radonangles)*np.pi/180,*params))**2))/np.ptp(circlipse(np.array(radonangles)*np.pi/180,*params))

        
        R_0 = params[0]
        R_1 = params[1]
        R_2 = params[2]
        phi0 = params[3]
        
        # see if fit is good
        ## check 1: enforce d_para > d_perp
        if np.abs(phi0)<70*np.pi/180:
            if best_circlipse[0]>best_circlipse[18]:
                print("failed requirement d_para > d_perp")
                raise ValueError('Fit not possible')
        else:
            if best_circlipse[0]<best_circlipse[18]: 
                print("failed requirement d_para > d_perp")
                raise ValueError('Fit not possible')
                
        ## check 2: preclude jumping between circlipses- max gap between consecutive points is 0.02M
        if np.max(np.abs(np.diff(best_circlipse)))>0.1*M_to_muas: 
            print("failed due to jumping between circlipses")
            raise ValueError('Fit not possible')
        

        d_1=2*(R_0+R_2)/M_to_muas
        d_2=2*(R_0+R_1)/M_to_muas
        d_perp=min(d_1, d_2)  ## d_perp <= d_para
        d_para=max(d_1, d_2)
        # fractional asymmetry
        f_a = 1-d_perp/d_para

        observables=[d_perp, d_para, f_a]


    except Exception as error:
        print("Fit not possible. The following exception occurred:")
        PrintException()
        best_circlipse=[np.nan]
        circlipses=[np.nan]
        g_circlipses=[np.nan]
        d_vals=[np.nan]
        g_vals=[np.nan]
        RMSD_a=np.nan
        RMSD_b=np.nan
        params=[np.nan, np.nan, np.nan, np.nan]
        observables=[np.nan, np.nan, np.nan]
        

    param_dict={
        "mu":best_circlipse,
        "circlipses":circlipses,
        "g_circlipses":g_circlipses,
        "d_vals":d_vals,
        "g_vals":g_vals,
        "rmsda":RMSD_a,
        "rmsdb":RMSD_b,
        "fit":params,
        "observables":observables,
        "power":power
    }
    
        # Create a directory for the results
    isExist = os.path.exists(path)
    if not isExist:
        os.makedirs(path)
    np.savez(path+"Fit_%s_%s_gfit_%s_mun_%s_sn_%s.npz"%(ulim1_fit,ulim2_fit, gfit_min, mu_n, s_n),**param_dict)


from aart import *

def gfactor(spins, i_angles, betaphi=1.0, betar=1.0, path='./Results/'):
    """
    :param spins: an iterable object containing the spins of the BH
    :param i_angles: an iterable object containing the inclination angles of the BH relative to the observer, in degrees
    :param betaphi: angular velocity
    :param betar: radial velocity
    :param path: path for saving the output. This should be consistent across the usage of all functions in this package    
    """
    for spin_case in spins:
        for i_case in i_angles:

            print("Computing the redshift factor at each point in the image plane \n")

            fnbands=path+"LensingBands_a_%s_i_%s.h5"%(spin_case,i_case)

            print("Reading file: ",fnbands)

            h5f = h5py.File(fnbands,'r')

            supergrid0=h5f['grid0'][:]
            mask0=h5f['mask0'][:]
            N0=int(h5f["N0"][0])

            supergrid1=h5f['grid1'][:]
            mask1=h5f['mask1'][:]
            N1=int(h5f["N1"][0])

            supergrid2=h5f['grid2'][:]
            mask2=h5f['mask2'][:]
            N2=int(h5f["N2"][0])

            h5f.close()

            fnbands=path+"Rays_a_%s_i_%s.h5"%(spin_case,i_case)

            print("Reading file: ",fnbands)

            h5f = h5py.File(fnbands,'r')

            rs0=h5f['rs0'][:]
            sign0=h5f['sign0'][:]
            #t0=h5f['t0'][:]
            phi0=h5f['phi0'][:]

            rs1=h5f['rs1'][:]
            sign1=h5f['sign1'][:]
            #t1=h5f['t1'][:]
            phi1=h5f['phi1'][:]

            rs2=h5f['rs2'][:]
            sign2=h5f['sign2'][:]
            #t2=h5f['t2'][:]
            phi2=h5f['phi2'][:]

            h5f.close()
            
            isco=rms(spin_case)
            
            thetao=i_case*np.pi/180

            i_g0 = obsint.gfactorf(supergrid0,mask0,sign0,spin_case,isco,rs0,phi0,thetao, betaphi, betar)
            i_g1 = obsint.gfactorf(supergrid1,mask1,sign1,spin_case,isco,rs1,phi1,thetao, betaphi, betar)
            i_g2 = obsint.gfactorf(supergrid2,mask2,sign2,spin_case,isco,rs2,phi2,thetao, betaphi, betar)

            i_g0 = (i_g0).reshape(N0,N0).T
            i_g1 = (i_g1).reshape(N1,N1).T
            i_g2 = (i_g2).reshape(N2,N2).T

            filename=path+"gfactors_a_%s_i_%s.h5"%(spin_case,i_case)

            h5f = h5py.File(filename, 'w')
            h5f.create_dataset('gs0', data=i_g0)
            h5f.create_dataset('gs1', data=i_g1)
            h5f.create_dataset('gs2', data=i_g2)

            h5f.close()
